/*
	joy2key
		This program polls the joystick device /dev/js0 and dispatches
		X events to a specific window based on joystick status.
		It is most useful for adding joystick support to X games that 
		don't nativly support the joystick, using joy2key to send
		XKeyPress messages in place of actually typing movement keys.
		This code is under the GNU Public License (see COPYING), 
		you can do whatever you want to it, and if you do anything really
		neat, I'd like to see it!

		- Peter Amstutz (amstpi@freenet.tlh.fl.us)

		special thanks to Zhivago of EFNet's #c for helping me out a LOT with
			the XSendMessage() code...

		The latest version of joy2key can be found in 
			ftp://sunsite.unc.edu/pub/Linux/X11/xutils
		or on joy2key's (rather spartan) website,
			http://www.freenet.tlh.fl.us/~amstpi/joy2key.html

		Revision History
		----------------
		1.0 (12 Aug 97) - First version.
		1.1 (04 Dec 97) - Thanks to Craig Lamparter (craiger@hemna.rose.hp.com), 
						joy2key now has support for four-buttoned 
						joysticks/gamepads!  Also some bugfixes.
*/

/* The standard joystick driver (current ver 0.8.0) only supports two buttons, 
but if you apply the patch joystick.patch (in this archive) to the joystick 
driver you can use four.  Then you can define this */
/* #define USE_FOUR_BUTTONS */

#define JOY2KEY_VERSION "1.1"

#define DEFAULT_X_THRESHOLD_HIGH 	700
#define DEFAULT_X_THRESHOLD_LOW 	300
#define DEFAULT_Y_THRESHOLD_HIGH 	700
#define DEFAULT_Y_THRESHOLD_LOW 	300
#define DEFAULT_UP_KEYSYM 			XK_Up
#define DEFAULT_DOWN_KEYSYM 		XK_Down
#define DEFAULT_LEFT_KEYSYM 		XK_Left
#define DEFAULT_RIGHT_KEYSYM 		XK_Right
#define DEFAULT_BUTTON1_KEYSYM		XK_z
#define DEFAULT_BUTTON2_KEYSYM		XK_a
#ifdef USE_FOUR_BUTTONS
#define DEFAULT_BUTTON3_KEYSYM		XK_s
#define DEFAULT_BUTTON4_KEYSYM		XK_x
#endif
#define DEFAULT_FREQUENCY			1000
#define DEFAULT_DEVICE				"/dev/js0"

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <linux/joystick.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <X11/keysym.h>
#include <time.h>

void main(int argc, char **argv)
{
	Display *thedisp;
	int thewindow=0, thescreen=0;
	FILE *foo;
	char string[255], winname[255], *device=DEFAULT_DEVICE;
	XEvent event;
	struct JS_DATA_TYPE js0;
	int i, jsfd,
		oldx, oldy, oldb,
                b1, b2, b3, b4, oldb1, oldb2, oldb3, oldb4,
		arup=DEFAULT_UP_KEYSYM, 
		ardown=DEFAULT_DOWN_KEYSYM, 
		arleft=DEFAULT_LEFT_KEYSYM, 
		arright=DEFAULT_RIGHT_KEYSYM, 
		but1=DEFAULT_BUTTON1_KEYSYM, 
		but2=DEFAULT_BUTTON2_KEYSYM,
#ifdef USE_FOUR_BUTTONS
		but3=DEFAULT_BUTTON3_KEYSYM, 
		but4=DEFAULT_BUTTON4_KEYSYM,
#endif
		xthreshhi=DEFAULT_X_THRESHOLD_HIGH,
		xthreshlow=DEFAULT_X_THRESHOLD_LOW,
		ythreshhi=DEFAULT_Y_THRESHOLD_HIGH,
		ythreshlow=DEFAULT_Y_THRESHOLD_LOW,
		waittime=DEFAULT_FREQUENCY;

	puts("joy2key - reads joystick status and dispatches X keyboard events");
	puts("By Peter Amstutz (amstpi@freenet.tlh.fl.us)");
	puts("This is free software under the GNU Public License");
	puts("              (see COPYING in the joy2key archive)");
	puts("You are welcome to use/modify this code, and please e-mail me");
	puts("if anything cool comes of it!");
	printf("Version: %s   Binary built on %s at %s\n\n", 
		JOY2KEY_VERSION, __DATE__, __TIME__);

	thedisp=XOpenDisplay(NULL);
	if(thedisp==NULL) 
	{
		printf("Error opening X Display");
		return;
	}
	thescreen=DefaultScreen(thedisp);

	if(argc<2) 
	{
		puts("Usage: joy2key \"Window Name\"");
		puts("\t[ -arrows [Up] [Down] [Left] [Right] ]");
		puts("\t[ -buttons12 [z] [a] ]");
		puts("\t[ -buttons34 [s] [x] ]");
		puts("\t[ -xthresh [hi[700]] [low[300]]");
		puts("\t[ -ythresh [hi[700]] [low[300]]");
		puts("\t[ -freq [1000] ]");
		puts("\t[ -dev [/dev/js0] ]");
		return;
	}

	sprintf(winname, "xwininfo -name \"%s\"", argv[1]);

	for(i=2; i<argc; i++)
	{
		if(!strcmp(argv[i], "-arrows"))
		{
			if(i+4>argc) 
			{
				puts("Not enough arguments to -arrows");
				return;
			}
			if((arup=XStringToKeysym(argv[++i]))==NoSymbol) printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
			if((ardown=XStringToKeysym(argv[++i]))==NoSymbol) printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
			if((arleft=XStringToKeysym(argv[++i]))==NoSymbol) printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
			if((arright=XStringToKeysym(argv[++i]))==NoSymbol) printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
			continue;
		}
		if(!strcmp(argv[i], "-buttons12"))
		{
			if(i+2>argc) 
			{
				puts("Not enough arguments to -buttons12");
				return;
			}
			if((but1=XStringToKeysym(argv[++i]))==NoSymbol) 
			{
				printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
				but1=DEFAULT_BUTTON1_KEYSYM;
			}
			if((but2=XStringToKeysym(argv[++i]))==NoSymbol) 
			{
				printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
				but2=DEFAULT_BUTTON2_KEYSYM;
			}
			continue;
		}
#ifdef USE_FOUR_BUTTONS
		if(!strcmp(argv[i], "-buttons34"))
		{
			if(i+2>argc) 
			{
				puts("Not enough arguments to -buttons34");
				return;
			}
			if((but3=XStringToKeysym(argv[++i]))==NoSymbol) 
			{
				printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
				but3=DEFAULT_BUTTON3_KEYSYM;
			}
			if((but4=XStringToKeysym(argv[++i]))==NoSymbol) 
			{
				printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
				but4=DEFAULT_BUTTON4_KEYSYM;
			}
			continue;
		}
#endif
		if(!strcmp(argv[i], "-xthresh"))
		{
			if(i+2>argc) 
			{
				puts("Not enough arguments to -xthresh");
				return;
			}
			xthreshhi=atoi(argv[++i]);
			xthreshlow=atoi(argv[++i]);
			continue;
		}
		if(!strcmp(argv[i], "-ythresh"))
		{
			if(i+2>argc) 
			{
				puts("Not enough arguments to -ythresh");
				return;
			}
			ythreshhi=atoi(argv[++i]);
			ythreshlow=atoi(argv[++i]);
			continue;
		}
		if(!strcmp(argv[i], "-freq"))
		{
			if(i+1>argc) 
			{
				puts("Not enough arguments to -freq");
				return;
			}
			if(atoi(argv[++i])==0) 
				puts("Invalid polling frequency, using default");
			else waittime=atoi(argv[++i]);
			continue;
		}
		if(!strcmp(argv[i], "-dev"))
		{
			if(i+1>argc) 
			{
				puts("Not enough arguments to -dev");
				return;
			}
			device=argv[++i];
			continue;
		}
		printf("Unknown option %s\n", argv[i]);
		return;
	}
	
	foo=popen(winname, "r");
	fgets(string, 255, foo);
	memset(string, 0, sizeof(string));
	fgets(string, 22, foo);
	fscanf(foo,"%x",&thewindow);
	pclose(foo);
	if(thewindow==0) 
	{
		puts("Error!  Can't find window to send events to!");
		return;
	}

	if((jsfd=open(device,O_RDONLY))==-1)
	{
		printf("Error opening %s!\n", device);
		puts("Are you sure you have joystick support in your kernel?");
		return;
	}
	memset(&js0, 0, sizeof(struct JS_DATA_TYPE));
	memset(&event, 0, sizeof(event));
	event.xkey.type=KeyPress;
	event.xkey.root=RootWindow(thedisp, thescreen);
	event.xkey.window=thewindow;
	event.xkey.subwindow=thewindow;
	event.xkey.same_screen=True;
	event.xkey.time=CurrentTime;

	arup=XKeysymToKeycode(thedisp, arup);
	ardown=XKeysymToKeycode(thedisp, ardown);
	arleft=XKeysymToKeycode(thedisp, arleft);
	arright=XKeysymToKeycode(thedisp, arright);
	but1=XKeysymToKeycode(thedisp, but1);
	but2=XKeysymToKeycode(thedisp, but2); 
#ifdef USE_FOUR_BUTTONS
	but3=XKeysymToKeycode(thedisp, but3); /* craiger */
	but4=XKeysymToKeycode(thedisp, but4);
#endif
	waittime=1000000/waittime;

	puts("Initialization complete, entering main loop, ^C to exit...");

	for(;;)
	{
		oldx=js0.x;
		oldy=js0.y;
		oldb=js0.buttons;
		oldb1=b1;
		oldb2=b2;
		oldb3=b3;
		oldb4=b4;
		read(jsfd, &js0, JS_RETURN);
                /* craiger mods, use 4 buttons, read simultaneous pressure */
                b1 = (js0.buttons & 0x01) >> 0;
                b2 = (js0.buttons & 0x02) >> 1;
#ifdef USE_FOUR_BUTTONS
                b3 = (js0.buttons & 0x04) >> 2;
                b4 = (js0.buttons & 0x08) >> 3;
#endif

		if(js0.x>xthreshhi && oldx<xthreshhi)
		{
			event.xkey.type=KeyPress;
			event.xkey.keycode=arright;
			event.xkey.display=thedisp;
			event.xkey.state=KeyPressMask;
			XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
		}
		if(js0.x<xthreshhi && oldx>xthreshhi)
		{
			event.xkey.type=KeyRelease;
			event.xkey.keycode=arright;
			event.xkey.display=thedisp;
			event.xkey.state=KeyReleaseMask;
			XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
		}
		if(js0.x<xthreshlow && oldx>xthreshlow)
		{
			event.xkey.type=KeyPress;
			event.xkey.keycode=arleft;
			event.xkey.display=thedisp;
			event.xkey.state=KeyPressMask;
			XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
		}
		if(js0.x>xthreshlow && oldx<xthreshlow)
		{
			event.xkey.type=KeyRelease;
			event.xkey.keycode=arleft;
			event.xkey.display=thedisp;
			event.xkey.state=KeyReleaseMask;
			XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
		}
		if(js0.y>ythreshhi && oldy<ythreshhi)
		{
			event.xkey.type=KeyPress;
			event.xkey.keycode=ardown;
			event.xkey.display=thedisp;
			event.xkey.state=KeyPressMask;
			XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
		}
		if(js0.y<ythreshhi && oldy>ythreshhi)
		{
			event.xkey.type=KeyRelease;
			event.xkey.keycode=ardown;
			event.xkey.display=thedisp;
			event.xkey.state=KeyReleaseMask;
			XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
		}
		if(js0.y<ythreshlow && oldy>ythreshlow)
		{
			event.xkey.type=KeyPress;
			event.xkey.keycode=arup;
			event.xkey.display=thedisp;
			event.xkey.state=KeyPressMask;
			XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
		}
		if(js0.y>ythreshlow && oldy<ythreshlow)
		{
			event.xkey.type=KeyRelease;
			event.xkey.keycode=arup;
			event.xkey.display=thedisp;
			event.xkey.state=KeyReleaseMask;
			XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
		}
		if(b1  && !(oldb1))
		{
				event.xkey.type=KeyPress;
    			event.xkey.keycode=but1;
				event.xkey.display=thedisp;
				event.xkey.state=KeyPressMask;
				XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
		}
		if(!(b1) && oldb)
		{
				event.xkey.type=KeyRelease;
    			event.xkey.keycode=but1;
				event.xkey.display=thedisp;
				event.xkey.state=KeyReleaseMask;
				XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
		}
		if(b2 && !(oldb2))                     
		{
				event.xkey.type=KeyPress;
    			event.xkey.keycode=but2;
				event.xkey.display=thedisp;
				event.xkey.state=KeyPressMask;
				XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
		}
		if(!(b2) && oldb2)
		{
				event.xkey.type=KeyRelease;
    			event.xkey.keycode=but2;
				event.xkey.display=thedisp;
				event.xkey.state=KeyReleaseMask;
				XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
		}
#ifdef USE_FOUR_BUTTONS
		if(b3 && !(oldb3))                     
		{
				event.xkey.type=KeyPress;
    			event.xkey.keycode=but3;
				event.xkey.display=thedisp;
				event.xkey.state=KeyPressMask;
				XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
		}
		if(!(b3) && oldb3)
		{
				event.xkey.type=KeyRelease;
    			event.xkey.keycode=but3;
				event.xkey.display=thedisp;
				event.xkey.state=KeyReleaseMask;
				XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
		}
		if(b4 && !(oldb4))                     
		{
				event.xkey.type=KeyPress;
    			event.xkey.keycode=but4;
				event.xkey.display=thedisp;
				event.xkey.state=KeyPressMask;
				XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
		}
		if(!(b4) && oldb4)
		{
				event.xkey.type=KeyRelease;
    			event.xkey.keycode=but4;
				event.xkey.display=thedisp;
				event.xkey.state=KeyReleaseMask;
				XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
		}
#endif

/* craiger */
/*                printf("Buttons: %d, %d, %d, %d.\n",b1,b2,b3,b4); */


		XFlush(thedisp);
		usleep(waittime);
	}

	XFlush(thedisp);
	XCloseDisplay(thedisp);

}
